/*
 * Created on Jun 30, 2005
 *
 * Copyright (c) 2005 Nooto Poject.  All rights reserved.
 */

package note.command;

import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import note.NoteApplication;
import note.control.NoteApplicationMode;
import note.model.Group;
import note.model.Page;
import note.model.TextGroup;
import note.viewer.GroupEditor;

/**
 * Class ModeChangeCommand
 * 
 * [hύXR}hB
 * 
 * @author Nooto Project Team
 * @version $Id: ModeChangeCommand.java,v 1.7 2006/02/06 20:12:31 camei Exp $
 */
public class ModeChangeCommand extends AbstractCommand {
	/***************************************************************************
	 * 萔
	 **************************************************************************/
	private static final long serialVersionUID = 1L;

	/***************************************************************************
	 * ϐ
	 **************************************************************************/
	private int mode = NoteApplicationMode.TEXT_MODE;

	// Undo
	private Page updatedPage = null;
	private TextGroup createdTextGroup = null;
	private GroupEditor fixedGroupEditor = null;
	private Group deletedGroup = null;
	private NoteApplicationMode oldMode = null;
	private NoteApplicationMode afterMode = null;

	/**
	 * RXgN^
	 * 
	 * @param mode ύX郂[h
	 * @param listener V[o[
	 * @see note.control.NoteApplicationMode
	 */
	public ModeChangeCommand(int mode, NoteApplication listener) {
		super(listener);
		this.mode = mode;
	}

	/**
	 * @return N[߂B
	 */
	public Object clone() {
		return new ModeChangeCommand(mode, getApplication());
	}

	/**
	 * @return ݎs\ǂ
	 * @see note.command.AbstractCommand#isEnabled()
	 */
	public boolean canExecute() {
		NoteApplicationMode currentMode = getMode();
		boolean canChangeTextMode = !currentMode.isTextMode()
				&& mode == NoteApplicationMode.TEXT_MODE;
		boolean canChangePictureMode = !currentMode.isPictureMode()
				&& mode == NoteApplicationMode.PICTURE_MODE;
		boolean canChangeGroupMode = !currentMode.isGroupEditMode()
				&& mode == NoteApplicationMode.GROUP_MODE;
		return canChangeTextMode || canChangePictureMode || canChangeGroupMode;
	}

	/**
	 * sB
	 */
	public void execute() {
		super.execute();

		// ݂̃[h擾
		NoteApplicationMode currentMode = getMode();
		oldMode = (NoteApplicationMode) currentMode.clone();// Undo̕ۑ

		// ͒̏ꍇ́A͒̃O[vm肷
		if (currentMode.isInputting()) {
			fixedGroupEditor = getEditableGroupEditor();// Undo̕ۑ
			fixedGroupEditor.setEditable(false);
		} else if (currentMode.isTextMode() && currentMode.isWaitInputting()) {// eLXgO[vꍇ
			// eLXgO[v폜
			updatedPage = getCurrentPage();
			deletedGroup = getEdittingTextGroup();
			GroupEditor editor = deletedGroup.getEditor();
			editor.setEditable(false);
			updatedPage.removeGroup(deletedGroup);
		}

		// [hύX
		currentMode.modeChanged(mode);

		if (currentMode.isTextMode()) {// eLXg̓[hɂȂꍇ
			// ̃eLXgO[vVK쐬
			updatedPage = getCurrentPage();
			createdTextGroup = createTextGroup();
		}

		afterMode = (NoteApplicationMode) currentMode.clone();// Undo̕ۑ
		repaint();// ĕ`
	}

	/***************************************************************************
	 * Undo Redo
	 **************************************************************************/
	/*
	 * @see AbstractCommand#isUndoableEdit()
	 */
	public boolean isUndoableEdit() {
		// \
		return true;
	}

	/*
	 * @see javax.swing.undo.AbstractUndoableEdit#getPresentationName()
	 */
	public String getPresentationName() {
		return "[h؂ւ";
	}

	/*
	 * @see javax.swing.undo.AbstractUndoableEdit#isSignificant()
	 */
	public boolean isSignificant() {
		return true;
	}

	/*
	 * @see javax.swing.undo.AbstractUndoableEdit#redo()
	 */
	public void redo() throws CannotRedoException {
		super.redo();

		// ݂̃[h擾
		NoteApplicationMode currentMode = getMode();

		// m肵O[vҏW\ɂ
		if (fixedGroupEditor != null) {
			fixedGroupEditor.setEditable(false);
		}

		// ̃eLXgO[v폜
		if (deletedGroup != null) {
			deletedGroup.getEditor().setEditable(false);
			updatedPage.removeGroup(deletedGroup);
		}

		// eLXgO[v𕜊
		if (createdTextGroup != null) {
			updatedPage.addGroup(createdTextGroup);
			createdTextGroup.getEditor().setEditable(true);
		}

		// ԑJڂ
		getMode().setValue(afterMode);

		repaint();// ĕ`
	}

	/*
	 * @see javax.swing.undo.AbstractUndoableEdit#undo()
	 */
	public void undo() throws CannotUndoException {
		super.undo();
		if (oldMode != null) {

			if (createdTextGroup != null) {
				// 쐬eLXgO[v폜
				updatedPage.removeGroup(createdTextGroup);
			}

			if (fixedGroupEditor != null) {
				// ÕO[vҏWԂɂ
				fixedGroupEditor.setEditable(true);

			} else if (deletedGroup != null) {
				// 폜O[vҏWԂɂ
				updatedPage.addGroup(deletedGroup);// ēxy[Wɒǉ
				deletedGroup.getEditor().setEditable(true);
			}

			// ԂtJڂ
			getMode().setValue(oldMode);

			repaint();// ĕ`
		}
	}

}
