/**
 * 
 */
package note.configuration;

import java.io.File;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import note.configuration.element.AbstractConfigurationElement;
import note.configuration.element.AbstractFunctionConfigurationElement;
import note.configuration.element.AnchorMoveFunctionConfigurationElement;
import note.configuration.element.FixGroupFunctionConfigurationElement;
import note.configuration.element.KeyAndFunctionConfigurationElement;
import note.configuration.element.KeyConfigurationElement;
import note.configuration.element.ModeChangeFunctionConfigurationElement;
import note.configuration.element.MoveKeyModifiersConfigurationElement;
import note.configuration.element.PageMoveFunctionConfigurationElement;
import note.configuration.element.PenChangeFunctionConfigurationElement;
import note.configuration.element.PictureInputFunctionConfigurationElement;
import note.configuration.element.PictureSelectCommandElement;
import note.configuration.element.TargetPicturePointMoveFunctionConfigurationElement;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * ݒXMLt@C͂A_x̃L[ݒɂp[TB
 * 
 * @author camei
 * @version $Id: KeyConfigurationXMLParser.java,v 1.2 2005/12/14 12:04:37 camei
 *          Exp $
 */
public class KeyConfigurationXMLParser {

	/***************************************************************************
	 * 萔
	 **************************************************************************/
	// ݒXMLt@CpX
	public static final String CONFIGURATION_PATH = "configuration/";// ݒtH_pX
	public static final String KEYMAP_PATH = CONFIGURATION_PATH + "";// L[ݒtH_pX

	public static final String ANCHOR_MOVE_KEY_CONFIGURATION_PATH = KEYMAP_PATH
			+ "anchormove-key.xml";
	public static final String FUNCTION_KEY_CONFIGURATION_PATH = KEYMAP_PATH
			+ "function-key.xml";
	public static final String PICTURE_INPUT_KEY_CONFIGURATION_PATH = KEYMAP_PATH
			+ "pictureinput-key.xml";
	public static final String MISC_KEY_CONFIGURATION_PATH = KEYMAP_PATH
			+ "misc-key.xml";
	public static final String[] KEY_CONFIGURATION_FILE_PATHS = {
			ANCHOR_MOVE_KEY_CONFIGURATION_PATH,
			FUNCTION_KEY_CONFIGURATION_PATH,
			PICTURE_INPUT_KEY_CONFIGURATION_PATH, MISC_KEY_CONFIGURATION_PATH };// L[ݒt@CpX

	// [gvf^O
	public static final String ANCHOR_MOVE_KEY_ROOT_TAG = "anchormove-key";
	public static final String PICTURE_INPUT_KEY_ROOT_TAG = "pictureinput-key";
	public static final String FUNCTION_KEY_ROOT_TAG = "function-key";
	public static final String MISC_KEY_ROOT_TAG = "misc-key";

	/***************************************************************************
	 * 
	 **************************************************************************/
	private ConceptualKeyConfiguration configuration;// _x̃L[ݒ

	/***************************************************************************
	 * \bh
	 **************************************************************************/
	/**
	 * L[ݒXMLt@C͂Ę_x̃L[ݒ߂
	 */
	public ConceptualKeyConfiguration parse() {

		// _x̃L[ݒ
		configuration = new ConceptualKeyConfiguration();

		// L[ݒXMLt@Cǂݍ݁A_x̃L[ݒݒ肷
		try {
			DocumentBuilder builder;

			String[] fileNames = KEY_CONFIGURATION_FILE_PATHS;// SẴL[ݒXMLt@C̃pX

			{// XMLp[T[p
				// hLgr_[t@Ng𐶐
				DocumentBuilderFactory dbfactory;
				dbfactory = DocumentBuilderFactory.newInstance();

				// hLgr_[𐶐
				builder = dbfactory.newDocumentBuilder();
			}

			// eݒXMLt@C͂
			for (int i = 0; i < fileNames.length; i++) {
				// p[XsDocumentIuWFNg擾
				Document doc = builder.parse(new File(fileNames[i]));// 

				// [gvf擾
				Element root = doc.getDocumentElement();

				// e[gvfɉ͂
				String rootTagName = root.getTagName();
				if (rootTagName.equals(ANCHOR_MOVE_KEY_ROOT_TAG)) {
					parseAnchorMoveKeyConfiguration(root);
				} else if (rootTagName.equals(PICTURE_INPUT_KEY_ROOT_TAG)) {
					parsePictureKeyConfiguration(root);
				} else if (rootTagName.equals(FUNCTION_KEY_ROOT_TAG)) {
					parseFunctionKeyConfiguration(root);
				} else if (rootTagName.equals(MISC_KEY_ROOT_TAG)) {
					parseMiscKeyConfiguration(root);
				}
				System.out.println("L[ݒǂݍ݊JnF" + root.getTagName());

			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		// ݒ̕ۑꂽݒ߂
		return configuration;
	}

	/***************************************************************************
	 * eXMLݒt@C̉͊֘A
	 **************************************************************************/

	/*
	 * AJ[ړL[̐ݒXMLt@C͂
	 */
	private void parseAnchorMoveKeyConfiguration(Element root) {

		parseAnchorMoveModifiers(root);// ړCqL[͂
		parseAnchorMoveKeyAndMoveDistance(root);// ړʏL[ƈړʂ̑Ή͂
	}

	/*
	 * ړCqL[͂
	 */
	private void parseAnchorMoveModifiers(Element root) {

		MoveKeyModifiersConfigurationElement modifiersConfigurationElement = new MoveKeyModifiersConfigurationElement();

		// ړCqL[͂
		String modifiersTagName = modifiersConfigurationElement.getTagName();
		NodeList modifierNodeList = root.getElementsByTagName(modifiersTagName);
		Element modifierElement = (Element) modifierNodeList.item(0);// ړCqL[vf
		modifiersConfigurationElement.setValue(modifierElement);

		// ړCqL[ۑ
		int keyCode = modifiersConfigurationElement.getKeyCode();
		String keyText = modifiersConfigurationElement.getKeyText();
		configuration.setMoveKeyModifiers(keyCode);
		configuration.setMoveKeyModifiersText(keyText);
	}

	/*
	 * ړʏL[ƈړʂ̑Ή͂
	 */
	private void parseAnchorMoveKeyAndMoveDistance(Element root) {

		List anchorMoveKeyConfiguration = configuration
				.getMoveKeyConfigurationList();// AJ[ړL[ݒ
		NodeList keyElementList = root
				.getElementsByTagName(KEY_CONFIGURATION_ELEMENT.getTagName());// AJ[ړL[^OvfɊ܂܂L[^Ovf̃Xg

		// ړL[ƈړʂ̑Ή͂āAݒɕۑ
		setKeyMapConfiguration(anchorMoveKeyConfiguration, keyElementList);
	}

	/*
	 * }`̓L[̐ݒXMLt@C͂
	 */
	private void parsePictureKeyConfiguration(Element root) {

		List pictureKeyConfiguration = configuration
				.getPictureKeyConfigurationList();// }`̓L[ݒ
		NodeList keyElementList = root
				.getElementsByTagName(KEY_CONFIGURATION_ELEMENT.getTagName());// }`̓L[^OvfɊ܂܂L[^Ovf̃Xg

		// }`̓L[Ɠ͂}`̑Ή͂āAݒɕۑ
		setKeyMapConfiguration(pictureKeyConfiguration, keyElementList);
	}

	/*
	 * t@NVL[̐ݒXMLt@C͂
	 */
	private void parseFunctionKeyConfiguration(Element root) {

		List functionKeyConfiguration = configuration
				.getFunctionKeyConfigurationList();// t@NVL[ݒ
		NodeList keyElementList = root
				.getElementsByTagName(KEY_CONFIGURATION_ELEMENT.getTagName());// t@NVL[^OvfɊ܂܂L[^Ovf̃Xg

		// t@NVL[Ǝs@\̑Ή͂āAݒɕۑ
		setKeyMapConfiguration(functionKeyConfiguration, keyElementList);
	}

	/*
	 * ̑̃L[̐ݒXMLt@C͂
	 */
	private void parseMiscKeyConfiguration(Element root) {

		List pictureEdittingKeyConfiguration = configuration
				.getPictureEdittingKeyConfigurationList();// }`ҏWL[ݒ
		NodeList keyElementList = root
				.getElementsByTagName(KEY_CONFIGURATION_ELEMENT.getTagName());// ̑L[^OvfɊ܂܂L[^Ovf̃Xg

		// }`ҏWL[Ǝs@\̑Ή͂āAݒɕۑ
		setKeyMapConfiguration(pictureEdittingKeyConfiguration, keyElementList);
	}

	/***************************************************************************
	 * ėp^Ovf͊֘A
	 **************************************************************************/

	// 萔
	private static final KeyConfigurationElement KEY_CONFIGURATION_ELEMENT = new KeyConfigurationElement();

	private static final AbstractConfigurationElement[] FC_ELEMENTS = {
			new AnchorMoveFunctionConfigurationElement(),
			new ModeChangeFunctionConfigurationElement(),
			new PageMoveFunctionConfigurationElement(),
			new PenChangeFunctionConfigurationElement(),
			new PictureInputFunctionConfigurationElement(),
			new TargetPicturePointMoveFunctionConfigurationElement(),
			new FixGroupFunctionConfigurationElement(),
			new PictureSelectCommandElement() };// SĂ̋@\ݒvf

	/*
	 * L[^Ovf̃Xg͂āAL[ݒvfƋ@\ݒvfw肳ꂽL[ݒɐݒ肷B
	 */
	private void setKeyMapConfiguration(List keyConfigurationList,
			NodeList keyElementList) {

		KeyConfigurationElement testKeyConfigurationElement = new KeyConfigurationElement();

		KeyConfigurationElement keyConfigurationElement;// L[ݒvf
		AbstractFunctionConfigurationElement functionConfigrationElement;// @\ݒvf

		// eL[vfɑΉtꂽ@\ݒ
		for (int i = 0; i < keyElementList.getLength(); i++) {
			Element keyElement = (Element) keyElementList.item(i);// L[^Ovf

			// L[ݒvf𐶐
			keyConfigurationElement = (KeyConfigurationElement) testKeyConfigurationElement
					.getNewInstance();
			keyConfigurationElement.setValue(keyElement);

			// L[^OvfΉtꂽ@\𐶐
			functionConfigrationElement = getFunctionConfiguraionElementByKeyElement(keyElement);

			// L[ݒɁAL[ݒvfƋ@\ݒvfo^
			KeyAndFunctionConfigurationElement keyAndFunction = new KeyAndFunctionConfigurationElement(
					keyConfigurationElement, functionConfigrationElement);
			keyConfigurationList.add(keyAndFunction);// ݒvfo^
		}
	}

	/*
	 * L[^Ovf͂AΉtĂ@\ݒvf߂
	 */
	private AbstractFunctionConfigurationElement getFunctionConfiguraionElementByKeyElement(
			Element keyElement) {

		NodeList functionElementList;// L[^OvfɊ܂܂@\^Ovf̃Xg
		Element functionElement;// L[^OvfɊ܂܂@\^Ovf
		AbstractFunctionConfigurationElement functionConfigurationElement = null;// @\ݒvf

		/*
		 * L[^OvfɑΉtꂽ@\^Ovf͂A @\^Ovfɉ@\ݒvf̃CX^X𐶐
		 */
		for (int i = 0; i < FC_ELEMENTS.length; i++) {
			// @\ݒvf̃^O^Ovf邩ׂ
			functionElementList = keyElement
					.getElementsByTagName(FC_ELEMENTS[i].getTagName());// @\^OvfXg

			if (functionElementList.getLength() != 0) {// ^Ovfꍇ
				// @\ݒvf𐶐
				functionElement = (Element) functionElementList.item(0);
				functionConfigurationElement = (AbstractFunctionConfigurationElement) FC_ELEMENTS[i]
						.getNewInstance();

				// @\^Ovf@\ݒvfɒlݒ肷
				functionConfigurationElement.setValue(functionElement);
				return functionConfigurationElement;
			}
		}

		return null;
	}
}
