package note.model;

import java.awt.Rectangle;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.Element;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;

import note.control.Pen;
import note.viewer.TextGroupPane;

/**
 * eLXgO[vB<br>
 * ꂽeLXg̃O[v\B<br>
 * 
 * @author camei
 * @version $Id: TextGroup.java,v 1.17 2006/01/31 19:31:51 camei Exp $
 */
public class TextGroup extends Group implements DocumentListener, Serializable {

	/***************************************************************************
	 * 萔
	 **************************************************************************/
	private static final long serialVersionUID = 1L;

	/***************************************************************************
	 * ϐ
	 **************************************************************************/
	private StyledDocument document;// eLXgO[v{́i́j
	private UndoableEventHandler undoableEditHandler = new UndoableEventHandler();

	/***************************************************************************
	 * \bh
	 **************************************************************************/

	/**
	 * RXgN^B
	 * 
	 * @param x ʒuxW
	 * @param y ʒuyW
	 */
	public TextGroup(int x, int y) {
		super(x, y);

		// Document 
		document = new DefaultStyledDocument();
		document.addDocumentListener(this);// hLgXi[ǉ
		document.addUndoableEditListener(undoableEditHandler);
	}

	/**
	 * RXgN^B
	 * 
	 * @param textGroup TextGroup
	 */
	private TextGroup(TextGroup textGroup) {
		super(textGroup.getX(), textGroup.getY());

		document = cloneDocument((DefaultStyledDocument) textGroup.document);
		document.addDocumentListener(this);// hLgXi[ǉ
		document.addUndoableEditListener(undoableEditHandler);
	}

	/**
	 * @return N[
	 */
	public Object clone() {
		return new TextGroup(this);
	}

	/*
	 * StyledDocument̕߂B
	 */
	private StyledDocument cloneDocument(DefaultStyledDocument document) {
		DefaultStyledDocument clone;

		// Document VACYĕ
		try {
			ByteArrayOutputStream stream = new ByteArrayOutputStream();
			ObjectOutput out = new ObjectOutputStream(stream);
			out.writeObject(document);
			ObjectInputStream in = new ObjectInputStream(
					new ByteArrayInputStream(stream.toByteArray()));
			clone = (DefaultStyledDocument) in.readObject();

		} catch (IOException e1) {
			e1.printStackTrace();
			throw new RuntimeException("DefaultStyledDocument̃VACYɎs");
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			throw new RuntimeException("DefaultStyledDocument̃VACYɎs");
		}

		// Document ̊̃XiO
		DocumentListener[] listeners = clone.getDocumentListeners();
		for (int i = 0; i < listeners.length; i++) {
			clone.removeDocumentListener(listeners[i]);
		}
		UndoableEditListener[] undoableListeners = clone
				.getUndoableEditListeners();
		for (int i = 0; i < undoableListeners.length; i++) {
			clone.removeUndoableEditListener(undoableListeners[i]);
		}

		return clone;
	}

	/***************************************************************************
	 * static \bh
	 **************************************************************************/
	/**
	 * y̑AttributeSetB
	 * 
	 * @param pen y
	 * @return yꂽ
	 */
	public static MutableAttributeSet getAttributeWithPen(Pen pen) {
		SimpleAttributeSet attribute = new SimpleAttributeSet();

		if (pen.isChangableSize()) {
			StyleConstants.setFontSize(attribute, pen.getFontSize());
		}
		if (pen.isChangableColor()) {
			StyleConstants.setForeground(attribute, pen.getColor());
		}
		if (pen.isChangableFontFamily()) {
			StyleConstants.setFontFamily(attribute, pen.getFontFamily());
		}

		return attribute;
	}

	/***************************************************************************
	 * 
	 **************************************************************************/
	/**
	 * ͂}B <br>
	 * 
	 * @param offset }ӏ
	 * @param text }镶
	 * @throws BadLocationException }ӏsȏꍇ
	 */
	public void insertString(int offset, String text)
			throws BadLocationException {
		AttributeSet attributeSet = new SimpleAttributeSet();
		insertString(offset, text, attributeSet);
	}

	/**
	 * ͂}B <br>
	 * 
	 * @param offset }ӏ
	 * @param text }镶
	 * @param attributeSet }镶͂̑
	 * @throws BadLocationException }ӏsȏꍇ
	 */
	public void insertString(int offset, String text, AttributeSet attributeSet)
			throws BadLocationException {
		document.insertString(offset, text, attributeSet);
	}

	/**
	 * eLXgvf}B
	 * 
	 * @param offset }ItZbg
	 * @param element }eLXgvf
	 * @throws BadLocationException ItZbg̎w肪sȏꍇ
	 */
	public void insertElement(int offset, TextElement element)
			throws BadLocationException {
		// TODO Ƒ}
		StyledDocument insertDocument = element.getDocument();

		String text;
		AttributeSet attributeSet;
		try {
			text = insertDocument.getText(0, insertDocument.getLength());
			attributeSet = insertDocument.getCharacterElement(0)
					.getAttributes();// b菈
		} catch (BadLocationException e) {
			e.printStackTrace();
			throw new Error("Document̃eLXg܂");
		}
		document.insertString(offset, text, attributeSet);

	}

	/**
	 * ͂폜
	 * 
	 * @param offset JnItZbg
	 * @param length 폜钷
	 * @throws BadLocationException ItZbgʒu̎w肪sȏꍇ̗O
	 */
	public void remove(int offset, int length) throws BadLocationException {
		try {
			document.remove(offset, length);
		} catch (BadLocationException e) {
			throw e;
		}
	}

	/**
	 * eLXg̑ςB
	 * 
	 * @param offset JnItZbg
	 * @param length ύX钷
	 * @param pen ύXɎgy
	 * @param replace ̑ύXŒu邩
	 * @throws BadLocationException ItZbgʒu̎w肪sȏꍇ̗O
	 */
	public void setCharacterAttributes(int offset, int length, Pen pen,
			boolean replace) throws BadLocationException {
		AttributeSet attributeSet = getAttributeWithPen(pen);

		document.setCharacterAttributes(offset, length, attributeSet, replace);

	}

	/***************************************************************************
	 * Group ̃I[o[Ch & 
	 **************************************************************************/

	public boolean intersects(int x, int y, int width, int height) {
		return ((TextGroupPane) getEditor()).intersects(x, y, width, height);
	}

	public boolean contains(int x, int y) {
		return ((TextGroupPane) getEditor()).contains(x, y);
	}

	public Rectangle getBounds() {
		return ((TextGroupPane) getEditor()).getBounds();
	}

	/***************************************************************************
	 * Cxg֘A
	 **************************************************************************/

	/**
	 * UndoableEditListenerǉB
	 * 
	 * @param listener UndoableEditListener
	 */
	public void addUndoableEditListener(UndoableEditListener listener) {
		// ܂铊
		this.undoableEditHandler.addUndoableEditListener(listener);
	}

	/**
	 * UndoableEditListener܂铊NXB
	 */
	private class UndoableEventHandler implements UndoableEditListener,
			Serializable {
		private transient UndoableEditListener listener;

		public void addUndoableEditListener(UndoableEditListener listener) {
			this.listener = listener;
		}

		public void undoableEditHappened(UndoableEditEvent e) {
			// ܂铊
			if (listener != null) {
				listener.undoableEditHappened(e);
			}
		}
	}

	/***************************************************************************
	 * DocumentListener ̎
	 **************************************************************************/
	/**
	 * insertUpdate ̎B
	 * 
	 * @param e DocumentEvent
	 */
	public void insertUpdate(DocumentEvent e) {
		update();
	}

	/**
	 * removeUpdate ̎B
	 * 
	 * @param e DocumentEvent
	 */
	public void removeUpdate(DocumentEvent e) {
		update();
	}

	/**
	 * changedUpdate ̎B
	 * 
	 * @param e DocumentEvent
	 */
	public void changedUpdate(DocumentEvent e) {
		update();
	}

	private void update() {
		// groupChangedListenerɃO[v̍XVʒm
		fireChangedUpdate();
	}

	/***************************************************************************
	 * Getter & Setter
	 **************************************************************************/
	/**
	 * TODO Ȃ
	 * 
	 * @return eLXgO[ṽGfB^
	 */
	public TextGroupPane getTextPane() {
		return (TextGroupPane) getEditor();
	}

	/**
	 * @return ̃O[v󂩂ǂ
	 */
	public boolean isEmpty() {
		return document.getLength() == 0;
	}

	/**
	 * @return Document
	 */
	public StyledDocument getDocument() {
		return document;
	}

}
