/*
 * Created on 2006/01/20
 *
 * Copyright (c) 2005 Nooto Poject.  All rights reserved.
 */
package note.viewer;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

import javax.swing.event.MouseInputAdapter;
import javax.swing.event.MouseInputListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoManager;
import javax.swing.undo.UndoableEdit;

import note.control.NoteApplicationMode;
import note.model.Group;
import note.model.Page;
import camei.select2.SelectElementHandler;
import camei.select2.SelectionManager;

/**
 * PageUIp̃}EXCxgnh Class PageMouseEventHandler
 * 
 * @see PageViewer
 * @author camei
 * @version $Id: PageMouseEventHandler.java,v 1.7 2006/02/06 20:12:36 camei Exp $
 */
public class PageMouseEventHandler extends MouseInputAdapter implements
		SelectionManager {

	/***************************************************************************
	 * 萔
	 **************************************************************************/
	private static final int RIGHT_BUTTON_MASK = MouseEvent.BUTTON3_MASK;// E{^}XN
	private static final int LEFT_BUTTON_MASK = MouseEvent.BUTTON1_MASK;// {^}XN

	/***************************************************************************
	 * ϐ
	 **************************************************************************/
	private PageViewer pageUI;
	private SelectElementHandler handler;
	private Rectangle selectionBounds = new Rectangle(0, 0, 0, 0);
	private List selectedGroupListTemporary;
	private UndoableEditListener undoableEditListener;// Undo\ȕҏWCxg郊Xi[
	static {
		new UndoManager();
	}

	/**
	 * RXgN^B
	 * 
	 * @param pageUI
	 */
	public PageMouseEventHandler(PageViewer pageUI,
			UndoableEditListener undoableEditListener) {
		this.pageUI = pageUI;
		this.handler = new SelectElementHandler(this);
		this.undoableEditListener = undoableEditListener;
	}

	/**
	 * TextGroupPane p̃}EXnhԂB
	 * 
	 * @param textGroupPane TextGroupPane
	 * @return
	 */
	public MouseInputListener getTextGroupPaneHandler(
			TextGroupPane textGroupPane) {
		return new TextGroupPaneHandler(textGroupPane);
	}

	/**
	 * TextGroupPanep̃}EXnhB
	 */
	public class TextGroupPaneHandler implements MouseInputListener {
		private TextGroupPane textGroupPane;

		public TextGroupPaneHandler(TextGroupPane textGroupPane) {
			this.textGroupPane = textGroupPane;
		}

		/*
		 * ̃eLXgO[vO[vI\ǂ߂B
		 */
		private boolean canSelectTextGroup() {
			// ҏW\łȂA܂͊ɑIĂꍇ
			return !textGroupPane.isEditable()
					|| pageUI.containsSelectedGroupList(textGroupPane
							.getGroup());
		}

		public void mouseDragged(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// I
				if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ

					int x = textGroupPane.getX() + e.getX();
					int y = textGroupPane.getY() + e.getY();
					handler.dragged(x, y);
					pageUI.repaint();
				}
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}

		public void mousePressed(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// I
				if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ
					startGroupEditMode();
					Rectangle bounds = textGroupPane.getBounds();
					int option = getSelectOption(e);
					handler.pressed(e.getX() + bounds.x, e.getY() + bounds.y,
							textGroupPane.getTextGroup(), option);
					pageUI.repaint();
				}
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();

			} else {
				// O[v̑I
				clearSelection();
				pageUI.repaint();
			}
		}

		public void mouseReleased(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// I
				if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ
					handler.released();
					pageUI.repaint();
				}
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}

		public void mouseClicked(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// I
				// IO[vENbNőIO[v폜
				if ((e.getModifiers() & RIGHT_BUTTON_MASK) != 0) {// E{^̏ꍇ
					if (isSelected(textGroupPane.getTextGroup())) {// IO[v|CgĂꍇ
						removeSelectedGroups();
						pageUI.repaint();
					}
				}
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}

		public void mouseEntered(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}

		public void mouseExited(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}

		public void mouseMoved(MouseEvent e) {
			if (canSelectTextGroup()) {// I\ȏꍇ
				// fBtHg̃}EXCxgs킹Ȃ
				e.consume();
			}
		}
	}

	/***************************************************************************
	 * MouseInputListener ̎
	 **************************************************************************/
	public void mousePressed(MouseEvent e) {
		if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ
			startGroupEditMode();
			handler.mousePressed(e);
			pageUI.repaint();
		}
	}

	public void mouseReleased(MouseEvent e) {
		if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ
			handler.mouseReleased(e);
			pageUI.repaint();
		}
	}

	public void mouseDragged(MouseEvent e) {
		if ((e.getModifiers() & LEFT_BUTTON_MASK) != 0) {// {^̏ꍇ
			handler.mouseDragged(e);
			pageUI.repaint();
		}
	}

	public void mouseClicked(MouseEvent e) {
		// IO[vENbNőIO[v폜
		if ((e.getModifiers() & RIGHT_BUTTON_MASK) != 0) {// E{^̏ꍇ
			Group clicked = (Group) getElement(e);
			if (isSelected(clicked)) {// IO[v|CgĂꍇ
				removeSelectedGroups();
				pageUI.repaint();
			}
		}
	}

	/***************************************************************************
	 * 
	 **************************************************************************/

	/**
	 * IĂO[v폜B TODO ɂׂłȂ\bhBPageViewerǂ
	 */
	public void removeSelectedGroups() {
		List targets = pageUI.getSelectedGroupList();// IĂO[vB

		// I
		clearSelection();

		// IĂO[v폜
		for (Iterator i = targets.iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			pageUI.getPage().removeGroup(group);
		}

		// UndoableEdit𓊂
		fireUndoableEdit(new GroupsDeleteEdit(targets));

		if (pageUI.getMode().isTextMode()) {// eLXg[h̏ꍇ
			// ̃eLXgO[v쐬
			Group group = pageUI.insertEmptyTextGroup();
			pageUI.getMode().groupUpdated(group);
		}
	}

	/*
	 * UndoableEditCxg𔭍sB
	 */
	private void fireUndoableEdit(UndoableEdit edit) {
		undoableEditListener.undoableEditHappened(new UndoableEditEvent(this,
				edit));
	}

	private void startGroupEditMode() {
		// ҏW̃O[vm
		GroupEditor editor = pageUI.getEditableGroupEditor();
		if (editor != null) {
			GroupSetEditableEdit edit = new GroupSetEditableEdit(editor, false);
			edit.execute();
			fireUndoableEdit(edit);
		}

		// O[v[hɏԑJ
		pageUI.getMode().modeChanged(NoteApplicationMode.GROUP_MODE);// O[vҏW[hɂ
	}

	/***************************************************************************
	 * SelectionManager ̎
	 **************************************************************************/
	public void select(Object element) {
		if (element instanceof Group) {
			Group group = (Group) element;
			// ̃eLXgO[v͑IȂ
			if (!group.isEmpty()) {// ̃eLXgO[vłȂꍇ

				pageUI.addSelectedGroup(group);
			}

			// TODO ɏ͂ȂCB
		} else if (element instanceof TextGroupPane) {
			Group group = ((TextGroupPane) element).getTextGroup();
			// ̃eLXgO[v͑IȂ
			if (!group.isEmpty()) {// ̃eLXgO[vłȂꍇ

				pageUI.addSelectedGroup(group);
			}
		}
		selectionUpdate();
	}

	public void unselect(Object element) {
		pageUI.removeSelectedGroup((Group) element);
		selectionUpdate();
	}

	public void clearSelection() {
		pageUI.clearSelectedGroups();
		selectionUpdate();
	}

	public void readySelectBounds() {
	}

	public void selectBounds(int x1, int y1, int x2, int y2) {
		setSelectionBounds(x1, y1, x2, y2);
		List selected = getGroupByBounds(selectionBounds);
		pageUI.clearSelectedGroups();
		pageUI.addSelectedGroups(selected);
	}

	public void decideSelectBoudns() {
		selectionUpdate();
	}

	public void readySelectBoundsAdditional() {
		// ݂̑IO[vL^
		selectedGroupListTemporary = pageUI.getSelectedGroupList();

		selectionUpdate();
	}

	public void selectBoundsAdditional(int x1, int y1, int x2, int y2) {
		setSelectionBounds(x1, y1, x2, y2);
		List inverser = getGroupByBounds(selectionBounds);
		Set inversed = inverseCollection(selectedGroupListTemporary, inverser);
		pageUI.clearSelectedGroups();
		pageUI.addSelectedGroups(inversed);
	}

	public void decideSelectBoundsAdditional() {
		selectedGroupListTemporary = null;
		selectionUpdate();
	}

	public void readyMoveSelection() {
	}

	/**
	 * I͈͂ړB
	 * 
	 * @param dx ړx
	 * @param dy ړy
	 * @see camei.select2.SelectionManager#moveSelection(int, int)
	 */
	public void moveSelection(int dx, int dy) {
		// IĂO[vړ
		for (Iterator i = pageUI.getSelectedGroupList().iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			group.translate(dx, dy);
		}
	}

	/**
	 * I͈͂̈ړm肷B
	 * 
	 * @see camei.select2.SelectionManager#decideMoveSelection()
	 */
	public void decideMoveSelection() {
		// UndoableEdit𔭍s
		Point distance = handler.getLatestDragDistance();
		int dx = (int) distance.getX();
		int dy = (int) distance.getY();
		fireUndoableEdit(new GroupMoveEdit(pageUI.getSelectedGroupList(), dx,
				dy));
	}

	/**
	 * I͈͂Rs[鏀B
	 * 
	 * @see camei.select2.SelectionManager#readyCopySelection()
	 */
	public void readyCopySelection() {
		List cloneList = new ArrayList();
		Page page = pageUI.getPage();

		// IĂO[v𕡐
		for (Iterator i = pageUI.getSelectedGroupList().iterator(); i.hasNext();) {
			Group group = (Group) i.next();

			// IĂO[v̕
			Group cloneGroup = (Group) group.clone();

			// y[Wɒǉ
			page.addGroup(cloneGroup);

			// ҏWs\ɂ
			cloneGroup.getEditor().setEditable(false);

			cloneList.add(cloneGroup);
		}

		// ݂̂IԂɂ
		pageUI.clearSelectedGroups();// I͈͂
		pageUI.addSelectedGroups(cloneList);
	}

	/**
	 * Rs[ړB
	 * 
	 * @param dx xړ
	 * @param dy yړ
	 */
	public void copySelection(int dx, int dy) {
		moveSelection(dx, dy);
	}

	/**
	 * Rs[ړm肷B
	 */
	public void decideCopySelection() {
		// UndoableEdit𔭍s
		List groupList = pageUI.getSelectedGroupList();
		fireUndoableEdit(new GroupCopyEdit(groupList));
	}

	/**
	 * IuWFNgҏWԂɂB
	 * 
	 * @see camei.select2.SelectionManager#edit(java.lang.Object)
	 */
	public void edit(Object element) {
		// I͈͂
		pageUI.clearSelectedGroups();

		// O[vҏW\Ԃɂ
		GroupEditor editor = ((Group) element).getEditor();
		editor.setEditable(true);
		UndoableEdit edit = new GroupSetEditableEdit(editor, true);
		fireUndoableEdit(edit);
	}

	public boolean isSelected(Object element) {
		return pageUI.containsSelectedGroupList(element);
	}

	/**
	 * |Cgɂvf߂B
	 * 
	 * @param e MouseEvent
	 * @return |Cgɂvf
	 */
	public Object getElement(MouseEvent e) {
		Page page = pageUI.getPage();
		List groupList = page.getGroupList();
		int x = e.getX();
		int y = e.getY();

		// |CgɂO[vT
		// ォ}ꂽO[v̂قAz-orderʂƂȂ悤ɈĂ
		for (ListIterator i = groupList.listIterator(groupList.size()); i
				.hasPrevious();) {
			Group group = (Group) i.previous();
			if (group.contains(x, y)) {
				return group;
			}
		}
		return null;
	}

	public int getSelectOption(MouseEvent e) {
		if (e.isControlDown()) {
			return SelectElementHandler.ADD;
		} else {
			return SelectElementHandler.SIMPLE;
		}
	}

	public Point getGlobalPoint(MouseEvent e) {
		return e.getPoint();
	}

	/*
	 * I͈͂XVꂽƂAĂяo鏈B
	 */
	private void selectionUpdate() {
		if (!pageUI.getSelectedGroupList().isEmpty()) {// IO[vꍇ
			startGroupEditMode();
		}
	}

	private void setSelectionBounds(int x1, int y1, int x2, int y2) {
		selectionBounds.x = (x1 < x2) ? x1 : x2;
		selectionBounds.y = (y1 < y2) ? y1 : y2;
		selectionBounds.width = (x1 < x2) ? x2 - x1 : x1 - x2;
		selectionBounds.height = (y1 < y2) ? y2 - y1 : y1 - y2;
	}

	private List getGroupByBounds(Rectangle bounds) {
		List groupList = pageUI.getPage().getGroupList();
		List result = new ArrayList();
		for (Iterator i = groupList.iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			if (group.intersects(bounds.x, bounds.y, bounds.width,
					bounds.height)
					&& !group.isEmpty()) {// I͈͋`̈ɌAłȂO[v̏ꍇ
				result.add(group);
			}
		}
		return result;
	}

	/*
	 * WWgċt]
	 */
	private Set inverseCollection(Collection c, Collection inverser) {
		Set result = new HashSet();

		// ϏW߂
		Collection intersection = new LinkedList(c);
		intersection.retainAll(inverser);

		// aWϏW
		result.addAll(c);
		result.addAll(inverser);
		result.removeAll(intersection);

		return result;
	}

	/***************************************************************************
	 * UndoableEdit TODO ƃt@C𕪂Ă
	 * execute\bh~BAbstractCommandNX̍Đ݌vB
	 **************************************************************************/
	private abstract class AbstractNoteUndoableEdit extends
			AbstractUndoableEdit {
		public abstract String getPresentationName();

		public String getUndoPresentationName() {
			return getPresentationName();
		}

		public String getRedoPresentationName() {
			return getPresentationName();
		}
	}

	private class GroupSetEditableEdit extends AbstractNoteUndoableEdit {
		private GroupEditor editor;
		private boolean isEditable = false;
		private GroupSetEditableEdit addEdit = null;
		private boolean isRemoved = false;

		public GroupSetEditableEdit(GroupEditor editor, boolean isEditable) {
			this.editor = editor;
			this.isEditable = isEditable;
		}

		public void execute() {
			// Undo̕ۑ
			Group group = editor.getGroup();
			if (!isEditable && group.isEmpty()) {// ̃O[vm肷ꍇ
				// O[v폜
				// TODO O[v̍폜PageViewerɂĂB߂B
				isRemoved = true;
			}

			editor.setEditable(isEditable);
			pageUI.repaint();
		}

		/*
		 * @see note.viewer.PageMouseEventHandler.AbstractNoteUndoableEdit#getPresentationName()
		 */
		public String getPresentationName() {
			String groupType = "";
			if (editor instanceof TextGroupPane) {
				groupType = "eLXgO[v";
			} else if (editor instanceof PictureGroupPane) {
				groupType = "}`O[v";
			}

			String editable = (isEditable) ? "ĕҏW" : "m";
			return groupType + editable;
		}

		/**
		 * m̍ĕҏW͂PXebvƂB
		 * 
		 * @see javax.swing.undo.AbstractUndoableEdit#addEdit(javax.swing.undo.UndoableEdit)
		 */
		public boolean addEdit(UndoableEdit anEdit) {
			if (anEdit instanceof GroupSetEditableEdit && !isEditable
					&& addEdit == null) {
				addEdit = (GroupSetEditableEdit) anEdit;
				return true;
			} else {
				return false;
			}
		}

		/*
		 * @see javax.swing.undo.AbstractUndoableEdit#redo()
		 */
		public void redo() throws CannotRedoException {
			super.redo();
			editor.setEditable(isEditable);
			if (!isEditable) {// m肵ꍇ
				// O[vҏW[hɂiËj
				pageUI.getMode().modeChanged(NoteApplicationMode.GROUP_MODE);// O[vҏW[hɂ
			}
			
			if (addEdit != null) {
				addEdit.redo();
			}

			pageUI.repaint();
		}

		/*
		 * @see javax.swing.undo.AbstractUndoableEdit#undo()
		 */
		public void undo() throws CannotUndoException {
			super.undo();
			if (addEdit != null) {
				addEdit.undo();
			}
			if (isRemoved) {// O[v폜ꍇ
				// O[v𕜊
				pageUI.getPage().addGroup(editor.getGroup());
			}
			editor.setEditable(!isEditable);
			pageUI.repaint();
		}
	}

	private class GroupsDeleteEdit extends AbstractNoteUndoableEdit {

		private Collection groupList;
		private GroupEditor editableGroupEditor;// 폜ꂽҏW̃O[v

		public GroupsDeleteEdit(Collection groupList) {
			this.groupList = groupList;
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				if (group.getEditor().isEditable()) {// 폜O[v̒ɕҏW̃O[vꍇ
					// 폜ꂽҏW̃O[vL^
					editableGroupEditor = group.getEditor();
					break;
				}
			}
		}

		public void undo() throws CannotUndoException {
			super.undo();

			// O[v𕜊
			Page page = pageUI.getPage();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				page.addGroup(group);
			}
			if (editableGroupEditor != null) {
				editableGroupEditor.setEditable(true);
			}
			pageUI.repaint();
		}

		public void redo() throws CannotRedoException {
			super.redo();

			// O[v폜
			Page page = pageUI.getPage();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				page.removeGroup(group);
			}

			pageUI.repaint();
		}

		public String getPresentationName() {
			return "O[v폜";
		}
	}

	private class GroupMoveEdit extends AbstractNoteUndoableEdit {

		private Collection groupList;
		private int dx;
		private int dy;

		public GroupMoveEdit(Collection groupList, int dx, int dy) {
			this.groupList = groupList;
			this.dx = dx;
			this.dy = dy;
		}

		public void undo() {
			super.undo();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				group.translate(-dx, -dy);
			}
			pageUI.repaint();
		}

		public void redo() {
			super.redo();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				group.translate(dx, dy);
			}
			pageUI.repaint();
		}

		public String getPresentationName() {
			return "O[vړ";
		}
	}

	private class GroupCopyEdit extends AbstractNoteUndoableEdit {

		private Collection groupList;

		public GroupCopyEdit(Collection groupList) {
			this.groupList = groupList;
		}

		public void undo() {
			super.undo();
			// O[v폜
			Page page = pageUI.getPage();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				page.removeGroup(group);
			}
			pageUI.repaint();
		}

		public void redo() {
			super.redo();
			// O[v𕜊
			Page page = pageUI.getPage();
			for (Iterator i = groupList.iterator(); i.hasNext();) {
				Group group = (Group) i.next();
				page.addGroup(group);
			}
			pageUI.repaint();
		}

		public String getPresentationName() {
			return "O[v";
		}

	}

	/***************************************************************************
	 * Getter
	 **************************************************************************/
	/**
	 * @return `Ï悪邩
	 */
	public boolean hasSelctionBounds() {
		return handler.isAddSelectingBounds() || handler.isSelectingBounds();
	}

	/**
	 * @return `Ï
	 */
	public Rectangle getSelectionBounds() {
		return this.selectionBounds;
	}

}
