/*
 * ViewerNX PageViewer.java
 */

package note.viewer;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.geom.Point2D;
import java.awt.print.PageFormat;
import java.awt.print.Pageable;
import java.awt.print.Printable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.event.MouseInputListener;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;

import note.NoteApplication;
import note.configuration.ApplicationConfiguration;
import note.control.NoteApplicationMode;
import note.model.Group;
import note.model.NoteStyle;
import note.model.Page;
import note.model.Picture;
import note.model.PictureGroup;
import note.model.TextGroup;

/**
 * y[W\R|[lgB
 * 
 * @author camei
 * @version $Id: PageViewer.java,v 1.40 2006/02/09 02:14:29 tajifu Exp $
 */
public class PageViewer extends JComponent implements Printable{

	/***************************************************************************
	 * 萔
	 **************************************************************************/
	private static final long serialVersionUID = 1L;
	private static final Insets ANCHOR_INSETS = new Insets(30, 30, 30, 30);// AJ[̕\`̈撆̈ړs\͈
	// private static final int MARGIN_SIZE = 20;

	// O[ṽnCCg
	private static final Stroke GROUP_HIGHLITER_STROKE = new BasicStroke(2.0f);
	private static final Color GROUP_HIGHLITER_COLOR = Color.GRAY;

	// p
	public static final int PAGE_EXISTS = 1;
	public static final int NO_SUCH_PAGE = 2;

	/***************************************************************************
	 * ϐ
	 **************************************************************************/
	private Page page;// \y[W

	// Fݒ
	private Color backgroundColor = Color.LIGHT_GRAY;// wiF

	// `֌W
	private Image backgroundImage;// ʉ摜
	private int imageX;// ʕ`XW
	private int imageY;// ʕ`YW

	// GUIi
	private AnchorViewer anchorViewer;
	private Map textGroupPanes = new HashMap();// TextGroupPaneB
	private Map pictureGroupPanes = new HashMap();// PictureGroupPaneB
	private GroupEditor editableGroupEditor = null;// ҏW̃O[vyC

	// Cxg֘A
	private UndoableEditListener undoableEditListener;// UndoableEdit󂯂郊Xi
	private KeyEventHandler keyEventHandler = new KeyEventHandler();
	private PageMouseEventHandler mouseEventHandler;
	private KeyListener keyListener;

	// control֘A
	private Point anchorPoint; // AJ[ʒu
	private NoteApplicationMode mode = new NoteApplicationMode();// ݂̃[h

	private List selectedGroupList = new LinkedList();// IĂO[vB
	private ApplicationConfiguration configuration;// ݒ

	// y[W̕\ʒuiy[WWň_j
	private int pageX = 0;
	private int pageY = 0;

	/***************************************************************************
	 * \bh
	 **************************************************************************/
	/**
	 * RXgN^B
	 * 
	 * @param page \y[W
	 * @param noteStyle m[gX^C
	 * @param configuration AvP[V̐ݒ
	 * @param undoableEditListener UndoableEditListener
	 */
	public PageViewer(Page page, NoteStyle noteStyle,
			ApplicationConfiguration configuration,
			UndoableEditListener undoableEditListener) {
		super();
		this.page = page;
		this.undoableEditListener = undoableEditListener;
		this.mouseEventHandler = new PageMouseEventHandler(this,
				undoableEditListener);
		this.configuration = configuration;
		this.keyListener = new MoveKeyCheckKeyListener();
		this.anchorPoint = new Point(noteStyle.getDefaultAnchorPoint());// AJ[̈ʒu

		// wi̗pӂ
		backgroundImage = noteStyle.createBackgroundImage();// y[W̔wi摜pӂ
		setBackground(backgroundColor);// R|[lg̔wiFw肷

		{// JComponent ̏
			// y[W̃TCYݒ
			Dimension size = new Dimension(noteStyle.getWidth(), noteStyle
					.getHeight());
			setPreferredSize(size);
			setMinimumSize(size);

			// Xi[̓o^
			addMouseListener(mouseEventHandler);
			addMouseMotionListener(mouseEventHandler);
			addKeyListener(keyListener);
			addKeyListener(new GroupDeleteKeyHandler());
			addFocusListener(new FocusHandler());

			// Cvbg\bh̓͂󂯕tȂݒɂi}`͒Ɂj
			enableInputMethods(false);
		}

		// AJ[\R|[lg
		anchorViewer = new AnchorViewer(anchorPoint, configuration
				.getCurrentPen(), mode);

		// y[W̃Xi[o^
		page.addPageChangedListener(this);

		// y[WɂO[vǂݍ
		List groupList = page.getGroupList();
		for (Iterator i = groupList.iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			groupAdded(group);
		}

		if (mode.isTextMode()) {// eLXg̓[hŊJnꍇ
			insertEmptyTextGroup();
		}

		// ĕ`
		repaint();
	}

	/*
	 * ̃eLXgO[vAJ[ʒuɑ}
	 */
	public TextGroup insertEmptyTextGroup() {
		TextGroup textGroup = page.createTextGroup(configuration
				.getCurrentPen(), undoableEditListener);// eLXgO[vVK쐬
		// undoableEditListener.undoableEditHappened(new UndoableEditEvent(this,
		// new EmptyTextGroupInsert(page, textGroup)));
		return textGroup;
	}

	/**
	 * ̃eLXgO[v쐬ҏWB TODO ͋ZB悭ȂB
	 */
	private class EmptyTextGroupInsert extends AbstractUndoableEdit {

		private Page page;
		private TextGroup createdTextGroup;
		private UndoableEdit beforeEdit;

		public EmptyTextGroupInsert(Page page, TextGroup createdTextGroup) {
			this.page = page;
			this.createdTextGroup = createdTextGroup;
		}

		/*
		 * @see javax.swing.undo.AbstractUndoableEdit#replaceEdit(javax.swing.undo.UndoableEdit)
		 */
		public boolean addEdit(UndoableEdit anEdit) {
			if (beforeEdit == null) {
				// OUndoableEditƘA
				beforeEdit = anEdit;
				return true;
			} else {
				return false;
			}
		}

		/*
		 * @see javax.swing.undo.AbstractUndoableEdit#redo()
		 */
		public void redo() throws CannotRedoException {
			super.redo();
			page.addGroup(createdTextGroup);
			if (beforeEdit != null) {
				beforeEdit.redo();
			}
		}

		/*
		 * @see javax.swing.undo.AbstractUndoableEdit#undo()
		 */
		public void undo() throws CannotUndoException {
			super.undo();
			if (beforeEdit != null) {
				beforeEdit.undo();
			}
			page.removeGroup(createdTextGroup);
		}

	}

	/***************************************************************************
	 * tH[JX֘A
	 **************************************************************************/

	private class FocusHandler implements FocusListener {
		// tH[JXƂ̓ݒ肷
		public void focusGained(FocusEvent e) {
			System.out.println(getClass().getName() + "#focusGained");

			NoteApplicationMode mode = page.getMode();
			if (mode.isTextMode() && editableGroupEditor != null
					&& editableGroupEditor.isEditable()) {// eLXg̓[h̏ꍇ
				if (!mode.isGroupEditMode()) {// O[vҏW[hłȂꍇ

					// eLXgyCɃtH[JX𓖂Ă
					((TextGroupPane) editableGroupEditor).requestFocus();

					System.out.println(getClass().getName()
							+ "    requestFocus->TextPane");
				}
			}
		}

		public void focusLost(FocusEvent e) {
			System.out.println(getClass().getName() + "#focusLost");
		}

	}

	/***************************************************************************
	 * 
	 **************************************************************************/

	/**
	 * ړCL[̉ĎL[XiB
	 */
	private class MoveKeyCheckKeyListener implements KeyListener {
		public void keyPressed(KeyEvent e) {
			int modifiers = PageViewer.this.configuration.getKeyConfiguration()
					.getMoveKeyModifiers();
			if (e.getKeyCode() == modifiers) {
				PageViewer.this.page.getMode().anchorMoveKeyPressed();
				repaint();
			}
		}

		public void keyReleased(KeyEvent e) {
			int modifiers = PageViewer.this.configuration.getKeyConfiguration()
					.getMoveKeyModifiers();
			if (e.getKeyCode() == modifiers) {
				PageViewer.this.page.getMode().anchorMoveKeyReleased();
				repaint();
			}
		}

		public void keyTyped(KeyEvent e) {
		}

	}

	/**
	 * AJ[|Cg̈ʒuw肵ʒuɈړB<br>
	 * TODO ҏW̃O[vS\悤ɂB
	 * 
	 * @param x AJ[ʒuxW
	 * @param y AJ[ʒuyW
	 */
	public void setAnchorLocation(int x, int y) {
		// AJ[|Cgy[W̊OɏoȂ悤ɂ(ݎgĂȂ)
		//Point2D includedPagePoint = getIncludedPagePoint(new Point(x, y));
		//x = (int) includedPagePoint.getX();
		//y = (int) includedPagePoint.getY();

		// AJ[|Cgړ
		anchorPoint.setLocation(x, y);

		// AJ[|Cgɉăy[WTCY
		resizeWithAnchorPoint();

		// AJ[|Cgɉăy[WXN[
		scrollWithAnchorPoint();
	}

	/**
	 * AJ[|Cg̈ʒuw肵ʒuɈړB<br>
	 * 
	 * @param point AJ[ʒu
	 */
	public void setAnchorLocation(Point point) {
		setAnchorLocation((int) point.getX(), (int) point.getY());
	}

	/***************************************************************************
	 * I֘A
	 **************************************************************************/

	/**
	 * @return IĂO[v邩
	 */
	public boolean hasSelection() {
		return !selectedGroupList.isEmpty();
	}

	/**
	 * IĂO[vǉB
	 * 
	 * @param group ǉO[v
	 */
	public void addSelectedGroup(Group group) {
		selectedGroupList.add(group);
	}

	/**
	 * IĂO[vBǉB
	 * 
	 * @param groups ǉO[vB
	 */
	public void addSelectedGroups(Collection groups) {
		selectedGroupList.addAll(groups);
	}

	/**
	 * IĂO[vO[vOB
	 * 
	 * @param group OO[v
	 * @return IĂO[vɎw肳ꂽO[v܂܂Ăꍇ<code>true</code>ԂB
	 */
	public boolean removeSelectedGroup(Group group) {
		return selectedGroupList.remove(group);
	}

	/**
	 * IĂO[vSẴO[vOB
	 */
	public void clearSelectedGroups() {
		selectedGroupList.clear();
	}

	/**
	 * @return IĂO[v
	 */
	public List getSelectedGroupList() {
		return new ArrayList(selectedGroupList);
	}

	/***************************************************************************
	 * `֘A
	 **************************************************************************/
	/**
	 * y[W̓e`悷B
	 * 
	 * @param g graphics context
	 * @see JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;

		// // y[W̕\ʒuɍW_킹
		// g2.translate(pageX, pageY);

		// ʂ`
		g2.drawImage(backgroundImage, imageX, imageY, null);

		// debug
		// g2.drawString(mode.toString(), 100, 20);// Ԃ\

		// O[ṽnCCg\
		for (Iterator i = page.getGroupList().iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			if (selectedGroupList.contains(group)) {// IĂꍇ
				Rectangle bounds = group.getBounds();
				g2.setStroke(GROUP_HIGHLITER_STROKE);
				g2.setPaint(GROUP_HIGHLITER_COLOR);
				g.drawRect(bounds.x, bounds.y, bounds.width, bounds.height);
			}
		}

		// I`̈`悷
		if (mouseEventHandler.hasSelctionBounds()) {
			g2.setStroke(new BasicStroke(1.0f));
			g2.setPaint(GROUP_HIGHLITER_COLOR);
			Rectangle selectionBounds = mouseEventHandler.getSelectionBounds();
			g.drawRect(selectionBounds.x, selectionBounds.y,
					selectionBounds.width, selectionBounds.height);
		}

		// }`O[vB`悷
		paintPictureGroupList(g2);

		// eLXgO[v`悷
		// eLXgO[v̓R|[lgȂ̂ŏɕ`悳
		super.paint(g);

		// AJ[`悷
		anchorViewer.paint(g2);
	}

	
	
	
	
	/*
	 * pBAJ[͕`ȂB
	 * 
	 */
	public void paintP(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;

		// ʂ`
		g2.drawImage(backgroundImage, imageX, imageY, null);

		// O[ṽnCCg\
		for (Iterator i = page.getGroupList().iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			if (selectedGroupList.contains(group)) {// IĂꍇ
				Rectangle bounds = group.getBounds();
				g2.setStroke(GROUP_HIGHLITER_STROKE);
				g2.setPaint(GROUP_HIGHLITER_COLOR);
				g.drawRect(bounds.x, bounds.y, bounds.width, bounds.height);
			}
		}

		// I`̈`悷
		if (mouseEventHandler.hasSelctionBounds()) {
			g2.setStroke(new BasicStroke(1.0f));
			g2.setPaint(GROUP_HIGHLITER_COLOR);
			Rectangle selectionBounds = mouseEventHandler.getSelectionBounds();
			g.drawRect(selectionBounds.x, selectionBounds.y,
					selectionBounds.width, selectionBounds.height);
		}

		// }`O[vB`悷
		paintPictureGroupList(g2);

		// eLXgO[v`悷
		// eLXgO[v̓R|[lgȂ̂ŏɕ`悳
		super.paint(g);

	}
			
	/*
	 * }`O[vB`悷B
	 */
	private void paintPictureGroupList(Graphics2D g2) {
		List pictureGroupList = page.getPictureGroupList();

		// A`GCAVOLɂ
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);

		// }`O[vB`悷
		for (int i = 0; i < pictureGroupList.size(); i++) {
			paintPictureGroup(g2, (PictureGroup) pictureGroupList.get(i));
		}

		// A`GCAVO𖳌ɂ
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_OFF);
	}

	/*
	 * }`O[v`悷B
	 */
	private void paintPictureGroup(Graphics2D g2, PictureGroup pictureGroup) {
		// O[v̌_ɕ`挴_ړ
		g2.translate(pictureGroup.getX(), pictureGroup.getY());

		((PictureGroupPane) pictureGroup.getEditor()).paint(g2);

		// `挴_ɖ߂
		g2.translate(-pictureGroup.getX(), -pictureGroup.getY());
	}

	/***************************************************************************
	 * CxgnhO֘A
	 **************************************************************************/
	private KeyListener applicationKeyListener;

	/**
	 * L[XiǉB
	 * 
	 * @param l KeyListener
	 */
	public void addKeyListener(KeyListener l) {
		// NoteApplication ւ̎QƂmۂB
		if (l instanceof NoteApplication) {
			applicationKeyListener = l;
		}
		super.addKeyListener(l);
	}

	/**
	 * TextGroupPane pL[Cxg̃nhB<br>
	 * L[CxgAvP[VɊۓB<br>
	 * TODO ĂBoĂB
	 */
	private class KeyEventHandler implements KeyListener {

		public void keyPressed(KeyEvent e) {
			if (!isEnableInputMethod()) {// ϊłȂ
				applicationKeyListener.keyPressed(e);
			}
		}

		public void keyReleased(KeyEvent e) {
			if (!isEnableInputMethod()) {// ϊłȂ
				applicationKeyListener.keyReleased(e);

			}
			if (mode.isGroupEditMode()) {// O[vҏW̏ꍇ

				// TODO n[hR[fBOłĂBƍ邱ƁB
				int keyCode = e.getKeyCode();
				if (keyCode == KeyEvent.VK_DELETE) {
					mouseEventHandler.removeSelectedGroups();
					repaint();
				}
			}
		}

		public void keyTyped(KeyEvent e) {
			if (!isEnableInputMethod()) {// ϊłȂ
				applicationKeyListener.keyTyped(e);
			}

			if (mode.isGroupEditMode()) {// O[vҏW̏ꍇ

				// eLXg͂Ȃ
				e.consume();
			}
		}

		private boolean isEnableInputMethod() {
			return mode.isTextMode()
					&& editableGroupEditor != null
					&& ((TextGroupPane) editableGroupEditor)
							.isEnableInputMethod();
		}
	}

	private class GroupDeleteKeyHandler extends KeyAdapter {
		public void keyReleased(KeyEvent e) {
			if (mode.isGroupEditMode()) {// O[vҏW̏ꍇ

				// TODO n[hR[fBOłĂBƍ邱ƁB
				int keyCode = e.getKeyCode();
				if (keyCode == KeyEvent.VK_DELETE) {
					mouseEventHandler.removeSelectedGroups();
					repaint();
				}
			}
		}
	}

	/***************************************************************************
	 * Cxg Page ̒ʒm
	 **************************************************************************/

	/**
	 * O[vǉꂽƂʒm
	 * 
	 * @param group ǉꂽO[v
	 */
	public void groupAdded(Group group) {
		if (group instanceof TextGroup) {
			TextGroupPane textGroupPane = (TextGroupPane) getGroupEditor(group);

			// y[WɃR|[lgƂĒǉ
			add(textGroupPane);

			// eLXgyCɃtH[JXĂ
			if (textGroupPane.isEditable()) {
				textGroupPane.requestFocus();
			}

			repaint();

		} else if (group instanceof PictureGroup) {
			// ̍s͕KvBȁB
			PictureGroupPane pictureGroupPane = (PictureGroupPane) getGroupEditor(group);
		} else {
			throw new IllegalArgumentException("O[v̎w肪s");
		}
	}

	/**
	 * O[v폜ꂽƂʒm
	 * 
	 * @param group 폜ꂽO[v
	 */
	public void groupRemoved(Group group) {
		if (editableGroupEditor != null
				&& group.getEditor() == editableGroupEditor) {// ҏWԂ̃O[v폜ꂽꍇ
			editableGroupEditor.setEditable(false);// O[vm肷
			editableGroupEditor = null;
			mode.eitableGroupRemoved();
		}

		if (group instanceof TextGroup) {// eLXgO[v̍폜w肳ꂽꍇ
			// TextGroupPane폜
			remove(getTextGroupPane((TextGroup) group));

			// tH[JX PageViewer Ɉڂ
			requestFocus();
		}
	}

	/*
	 * TextGroupPane ̃R|[lgɓo^B
	 */
	private void registerTextGroupPane(TextGroupPane textGroupPane) {

		// edittingTextPane.setInputMap(0,null);
		textGroupPane.addKeyListener(keyEventHandler);
		textGroupPane.addKeyListener(keyListener);

		// eLXgyCɃXi[ǉ
		MouseInputListener inputListener = mouseEventHandler
				.getTextGroupPaneHandler(textGroupPane);
		textGroupPane.addMouseListener(inputListener);
		textGroupPane.addMouseMotionListener(inputListener);
	}

	private void registerPictureGroupPane(PictureGroupPane pictureGorupPane) {

	}

	/***************************************************************************
	 * Cxg GroupEditor ̒ʒm<br>
	 **************************************************************************/
	/**
	 * O[vҏW\ԂXVꂽƂʒmB
	 * 
	 * @param groupEditor GroupEditor
	 */
	public void groupEditableChanged(GroupEditor groupEditor) {
		if (groupEditor.isEditable()) {// ҏW\ԂɂȂꍇ

			// ݂̕ҏW\O[vҏWs\ɂ
			if (editableGroupEditor != null && editableGroupEditor != groupEditor) {
				editableGroupEditor.setEditable(false);
			}

			// ҏW\O[vݒ肷
			editableGroupEditor = groupEditor;

			// AJ[ʒuXV
			groupAnchorPointMoved(groupEditor);

			// O[v̑I
			clearSelectedGroups();

			// ԑJ
			mode.groupSetEditable(groupEditor.getGroup());

		} else {// ҏWs\ԂɂȂꍇ
			if (editableGroupEditor == groupEditor) {// ҏW̃O[vҏWs\ɂȂꍇ
				editableGroupEditor = null;
				mode.groupFixed();
			}

			Group group = groupEditor.getGroup();
			if (group.isEmpty()) {// ̃O[vҏWs\ɂȂꍇ
				// ̃O[v폜
				page.removeGroup(group);
			}

			requestFocus();
		}

		repaint();
	}

	/**
	 * AJ[ʒmB
	 */
	public void groupAnchorPointMoved(GroupEditor groupEditor) {
		if (editableGroupEditor != null && editableGroupEditor == groupEditor) {
			// AJ[̈ʒuXV
			updateAnchorPoint();

			// ĕ`悷
			repaint();
		}
	}

	/**
	 * O[vύXXVꂽƂʒmB
	 * 
	 * @param groupEditor
	 */
	public void groupChanged(GroupEditor groupEditor) {
		Group group = groupEditor.getGroup();
		mode.groupUpdated(group);
		repaint();
	}

	/**
	 * O[vړʒmB
	 * 
	 * @param groupEditor
	 */
	public void groupMoved(GroupEditor groupEditor) {
		if (editableGroupEditor != null && editableGroupEditor == groupEditor) {
			groupAnchorPointMoved(groupEditor);
		}
	}

	/*
	 * AJ[̈ʒuXVB
	 */
	private void updateAnchorPoint() {
		// O[vWy[WWɕϊ
		Point localAnchor = editableGroupEditor.getAnchorPoint();
		Group group = editableGroupEditor.getGroup();
		localAnchor.translate(group.getX(), group.getY());

		// AJ[̈ʒuݒ肷
		setAnchorLocation(localAnchor);
	}

	/*
	 * AJ[|Cg̈ʒuɉăXN[
	 */
	private void scrollWithAnchorPoint() {
		int x = (int) anchorPoint.getX();// AJ[xW
		int y = (int) anchorPoint.getY();// AJ[yW

		// \Ă`̈擾
		Rectangle visible = getVisibleRect();

		// AJ[ViewPort͂ݏoȂ悤ɉ̈ړ
		if (visible.x + ANCHOR_INSETS.left > x) {
			visible.x = x - ANCHOR_INSETS.left;
		} else if (visible.x + visible.width - ANCHOR_INSETS.right < x) {
			visible.x = x + ANCHOR_INSETS.right - visible.width;
		}
		if (visible.y + ANCHOR_INSETS.top > y) {
			visible.y = y - ANCHOR_INSETS.top;
		} else if (visible.y + visible.height - ANCHOR_INSETS.bottom < y) {
			visible.y = y + ANCHOR_INSETS.bottom - visible.height;
		}

		// V`̈ݒ
		scrollRectToVisible(visible);
	}

	/*
	 * AJ[|Cg̈ʒuɉāCTCY FIXME AJ[𓮂1.AJ[ 2.O[ṽAJ[ B
	 * ܂1x̃AJ[ړɑ΂AĂ2x̃\bh͎sĂB
	 * 2xsƓ͐킾A1x̏ꍇAɓ삵ȂiTCỸ^C~OsK؁jB
	 * R͕ȂA܂Ă̂łقĂB
	 */
	private void resizeWithAnchorPoint() {
		int x = (int) anchorPoint.getX();// AJ[xW
		int y = (int) anchorPoint.getY();// AJ[yW
		int width = getWidth();// ݂̃TCY
		int height = getHeight();// ݂̃TCY
		int addTop = 0;
		int addLeft = 0;
		int addRight = 0;
		int addBottom = 0;
		boolean isNeedResize = false;

		// ǉTCY߂
		if (x + ANCHOR_INSETS.right > width) {// Eɂ͂ݏoꍇ
			addRight = x + ANCHOR_INSETS.right - width;
			isNeedResize = true;
		} else if (x - ANCHOR_INSETS.left < 0) {// ɂ͂ݏoꍇ
			addLeft = -(x - ANCHOR_INSETS.left);
			isNeedResize = true;
		}

		if (y + ANCHOR_INSETS.bottom > height) {// ɂ͂ݏoꍇ
			addBottom = y + ANCHOR_INSETS.bottom - height;
			isNeedResize = true;
		} else if (y - ANCHOR_INSETS.top < 0) {// ɂ͂ݏoꍇ
			addTop = -(y - ANCHOR_INSETS.top);
			isNeedResize = true;
		}

		// TCY
		if (isNeedResize) {
			addSize(addTop, addLeft, addBottom, addRight);
		}
	}

	/*
	 * wɕtăy[WTCYB
	 */
	private void addSize(int top, int left, int bottom, int right) {
		int newWidth = getWidth() + left + right;
		int newHeight = getHeight() + top + bottom;

		// R|[lgTCY
		Dimension newSize = new Dimension(newWidth, newHeight);
		setPreferredSize(newSize);
		setSize(newSize);
		
		System.out.println("newSize:" + newSize);
		System.out.println("size:" + getSize());

		// eړ
		if (top != 0 || left != 0) {
			moveContents(left, top);
		}
	}

	/*
	 * ʂɕ\Ăe܂Ƃ߂Ĉړ
	 */
	private void moveContents(int distanceX, int distanceY) {
		pageX += distanceX;
		pageY += distanceY;

		// ʉ摜ړ
		imageX += distanceX;
		imageY += distanceY;

		// AJ[ړ
		anchorPoint.translate(distanceX, distanceY);

		// O[vړ
		List groupList = page.getGroupList();
		for (Iterator i = groupList.iterator(); i.hasNext();) {
			Group group = (Group) i.next();
			group.translate(distanceX, distanceY);
		}
	}

	/*
	 * y[W̑΍W󂯎āAy[W͂ݏoĂ璆Ɏ܂_Ԃ
	 * 
	 * @param point y[W̑΍W @return y[WɎ܂邱Ƃۏ؂ꂽAy[W̑΍W
	 */
	private Point2D getIncludedPagePoint(Point point) {
		int x = 0;
		int y = 0;

		// 󂯎W͂ݏoĂy[WɎ߂
		if (point.getX() < getX()) {
			x = getX();
		} else if (x > getX() + getHeight()) {
			x = getX() + getHeight();
		}
		if (point.getY() < getY()) {
			y = getY();
		} else if (point.getY() > getY() + getHeight()) {
			y = getY() + getHeight();
		}
		return new Point(x, y);
	}

	/***************************************************************************
	 * GroupEditor ֘A
	 **************************************************************************/
	/*
	 * O[vɑΉ GroupEditor 擾
	 */
	private GroupEditor getGroupEditor(Group group) {
		if (group instanceof TextGroup) {
			return getTextGroupPane((TextGroup) group);
		} else if (group instanceof PictureGroup) {
			return getPictureGroupPane((PictureGroup) group);
		} else {
			throw new RuntimeException("O[v̎w肪s");
		}
	}

	/***************************************************************************
	 * TextGroupPane ֘A
	 **************************************************************************/

	private TextGroupPane getTextGroupPane(TextGroup textGroup) {
		String key = getTextGroupKey(textGroup);
		if (textGroupPanes.containsKey(key)) {
			return (TextGroupPane) textGroupPanes.get(key);
		} else {
			return createTextGroupPane(textGroup);
		}
	}

	private TextGroupPane createTextGroupPane(TextGroup textGroup) {
		String key = getTextGroupKey(textGroup);
		TextGroupPane newPane = new TextGroupPane(textGroup, this,
				configuration, undoableEditListener);
		textGroupPanes.put(key, newPane);
		registerTextGroupPane(newPane);
		return newPane;
	}

	private String getTextGroupKey(TextGroup textGroup) {
		return "TextGroup" + textGroup.hashCode();
	}

	/***************************************************************************
	 * PictureGroupPane ֘A
	 **************************************************************************/
	private PictureGroupPane getPictureGroupPane(PictureGroup pictureGroup) {
		String key = "pictureGroup" + pictureGroup.hashCode();
		if (pictureGroupPanes.containsKey(key)) {
			return (PictureGroupPane) pictureGroupPanes.get(key);
		} else {
			return createPictureGroupPane(pictureGroup);
		}
	}

	private PictureGroupPane createPictureGroupPane(PictureGroup pictureGroup) {
		Object key = getGroupKey(pictureGroup);
		PictureGroupPane newPane = new PictureGroupPane(pictureGroup, this);
		pictureGroupPanes.put(key, newPane);
		registerPictureGroupPane(newPane);
		return newPane;
	}

	private String getGroupKey(Group group) {
		return "group" + group.hashCode();
	}

	/***************************************************************************
	 * 
	 **************************************************************************/
	/*
	 * B
	 */
	public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
		int x = (int) pageFormat.getImageableX();
		int y = (int) pageFormat.getImageableY();
		g.translate(x, y);
		if (pageIndex == 0) {
			paintP(g);
			return Printable.PAGE_EXISTS;
		} else {
			return Printable.NO_SUCH_PAGE;
		}
	}

	/***************************************************************************
	 * Getter & Setter
	 **************************************************************************/
	public boolean containsSelectedGroupList(Object element) {
		return selectedGroupList.contains(element);
	}

	public Page getPage() {
		return page;
	}

	public Point getAnchorPoint() {
		return new Point(anchorPoint);
	}

	public NoteApplicationMode getMode() {
		return mode;
	}

	public GroupEditor getEditableGroupEditor() {
		return this.editableGroupEditor;
	}

	public TextGroupPane getEditableTextGroupPane() {
		return (TextGroupPane) editableGroupEditor;
	}

	public PictureGroupPane getEditablePictureGroupPane() {
		return (PictureGroupPane) editableGroupEditor;
	}
	
}
