/**
 * 
 */
package note.viewer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.InputEvent;
import java.awt.event.InputMethodEvent;
import java.awt.event.InputMethodListener;
import java.awt.event.KeyEvent;

import javax.swing.InputMap;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditListener;
import javax.swing.plaf.basic.BasicTextPaneUI;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultCaret;
import javax.swing.text.DefaultEditorKit;
import javax.swing.text.JTextComponent;
import javax.swing.text.Keymap;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.StyledDocument;
import javax.swing.text.JTextComponent.KeyBinding;

import note.configuration.ApplicationConfiguration;
import note.control.Pen;
import note.model.Group;
import note.model.GroupChangedListener;
import note.model.NoteElement;
import note.model.TextGroup;

/**
 * @author Administrator
 * @version $Id: TextGroupPane.java,v 1.17 2006/02/08 20:02:02 camei Exp $
 */
public class TextGroupPane extends JTextPane implements GroupChangedListener,
		CaretListener, GroupEditor {

	/***************************************************************************
	 * 萔
	 **************************************************************************/
	private static final long serialVersionUID = 1L;
	private static final Insets EDITOR_INSETS = new Insets(0, 3, 0, 3);// GfB^̃}[W

	// ɂV[gJbgL[
	private static final KeyStroke MENU_KEY = KeyStroke.getKeyStroke(
			KeyEvent.VK_F10, 0);
	private static final KeyStroke BACK_SPACE = KeyStroke.getKeyStroke("ctrl H");
//	private static final KeyStroke PASTE_KEY = KeyStroke.getKeyStroke(KeyEvent.VK_V, InputEvent.CTRL_MASK);
	private static final KeyStroke[] REMOVED_KEYS = { /*UNDO_KEY, REDO_KEY,*/
			MENU_KEY, /*PASTE_KEY,*/ BACK_SPACE };

	/***************************************************************************
	 * ϐ
	 **************************************************************************/
	private TextGroup textGroup;
	private Pen currentPen;
	private ApplicationConfiguration configuration;
	private PageViewer pageUI;
	private Point localAnchorPoint = new Point(0, 0);// O[vAJ[ʒu

	/**
	 * RXgN^B
	 * 
	 * @param textGroup eLXgO[v
	 */
	public TextGroupPane(TextGroup textGroup, PageViewer pageUI,
			ApplicationConfiguration configuration,
			UndoableEditListener undoableEditListener) {
		super(textGroup.getDocument());
		setEditable(false);
		this.textGroup = textGroup;
		this.pageUI = pageUI;
		this.configuration = configuration;
		this.currentPen = configuration.getCurrentPen();

		// Document ̏ݒ肷
		StyledDocument document = textGroup.getDocument();
		document.setParagraphAttributes(0, 0, TextGroup
				.getAttributeWithPen(currentPen), true);

		// UIݒ肷
		setUI(new TextGroupPaneUI());

		// R|[lgݒ肷
		setMargin(EDITOR_INSETS);// }[Wݒ
		setSuitableSize();// KȃTCYɂ
		setLocationToFirstLineBottom();// ʒuݒ肷
		setOpaque(false);// wi𓧖ɂ

		// Lbgݒ肷
		// ߂BȂ񂩁B
		// setCaret(new TextGroupCaret(currentPen));

		{// Xi[̐ݒ
			textGroup.setEditor(this);

			// eLXgO[ṽXi[
			// TODO ƃVACŶƂɂ܂ŏĂ̂łłȂ
			textGroup.addGroupChangedListener(this);
			textGroup.addUndoableEditListener(undoableEditListener);

			// Document ̃Xi[
			document.addDocumentListener(new DocumentAttributesChanger());// ȕϊɓ͂ꂽeLXg̑ύX

			// ̃R|[lg̃Xi[
			addInputMethodListener(new InputMethodWactcher());// ȕϊĎ
			addFocusListener(new TextGroupPaneFocusListener());
			addCaretListener(this);
		}

		{// InputMap 
			// JTextComponent ssȃV[gJbgL[폜
//			Keymap keymap = getKeymap();
//			for (int i = 0; i < REMOVED_KEYS.length; i++) {
//				keymap.removeKeyStrokeBinding(REMOVED_KEYS[i]);
//			}
			
			KeyBinding[] bindings = new KeyBinding[REMOVED_KEYS.length];
			Keymap keymap = getKeymap();
			for (int i = 0; i < REMOVED_KEYS.length; i++) {
				bindings[i] = new KeyBinding(REMOVED_KEYS[i], DefaultEditorKit.defaultKeyTypedAction);
			}
			JTextComponent.loadKeymap(keymap, bindings, this.getActions()); 
//			setKeymap(keymap);
			
//			InputMap inputMap = getInputMap();
//			for(int i = 0; i < REMOVED_KEYS.length; i++) {
//				inputMap.remove(REMOVED_KEYS[i]);
//			}
//			setInputMap(WHEN_FOCUSED, inputMap);
		}

		// java vefr 1.4 ȉ̏ꍇAł̃R[hȂƔwiɂȂȂB
		// TODO łŏ̂ЂƂ߂͓ɂȂȂB
		setOpaque(false);// wi𓧖ɂ
	}

	/**
	 * tH[JX郊XiB
	 */
	private class TextGroupPaneFocusListener implements FocusListener {
		/**
		 * tH[JX̏B
		 * 
		 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
		 */
		public void focusGained(FocusEvent e) {
			System.out.println(getClass().getName() + "[" + this.hashCode()
					+ "]" + "focusGained");
			if (!isEditable()) {// ҏWԂłȂꍇ
				// y[WɃtH[JXn
				pageUI.requestFocus();
				System.out.println("   focus->" + pageUI.getClass().getName());
			}
		}

		/**
		 * tH[JXȂȂƂ̏B
		 * 
		 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
		 */
		public void focusLost(FocusEvent e) {
		}
	}

	/***************************************************************************
	 * ͊֘A
	 **************************************************************************/

	/**
	 * eLXg̑ύX
	 * 
	 * @param offset ύXJnʒu
	 * @param length ύXeLXg̒
	 * @param pen ύXɎgy
	 * @param replace ɂ鑮u邩
	 * @throws BadLocationException
	 */
	public void setCharacterAttributes(int offset, int length, Pen pen,
			boolean replace) throws BadLocationException {

		textGroup.setCharacterAttributes(offset, length, pen, replace);
	}

	/**
	 * I͈͂̕ύX
	 * 
	 * @param pen y
	 */
	public void setSelectionAttributes(Pen pen, boolean replace) {
		int beginOffset = getSelectionStart();
		int length = getSelectionEnd() - beginOffset;
		try {
			setCharacterAttributes(beginOffset, length, pen, replace);
		} catch (BadLocationException e) {
			throw new RuntimeException("I͈͂̃ItZbgs");
		}
	}

	/**
	 * IĂeLXg܂ދ`̈߂B <br>
	 * 
	 * @return O[vW`Ԃ̋`BI͈͂ꍇ<code>null</code>߂B
	 */
	public Rectangle getSelectionBounds() {

		// I͈͂ĂȂꍇnullԂ
		if (!hasSelection()) {
			return null;
		}

		int beginOffset = getSelectionStart();
		int endOffset = getSelectionEnd();

		try {
			return getTextBounds(beginOffset, endOffset);
		} catch (BadLocationException e) {
			throw new RuntimeException("I͈͂̃ItZbgʒus");
		}

	}

	/**
	 * @param beginOffset eLXg̊JnItZbg
	 * @param endOffset eLXgIItZbg
	 * @return w肳ꂽeLXg܂ދ`̈
	 * @throws BadLocationException ItZbg̎w肪sȏꍇ
	 */
	public Rectangle getTextBounds(int beginOffset, int endOffset)
			throws BadLocationException {

		// I͈͂܂ދ`̈
		Rectangle beginRectangle = modelToView(beginOffset);

		int left = beginRectangle.x;// `̈̍xW
		int top = beginRectangle.y;// `̈̏yW
		int right = 0;// `̈̉ExW
		int bottom = 0;// `̈̉yW

		// I͈͂̉ʒuSĒׂāASĂ܂ޔ͈͂vZ
		Rectangle bounds;// hbg̉ʒu
		for (int i = beginOffset; i <= endOffset; i++) {
			bounds = modelToView(i);
			if (left > bounds.x) {
				left = bounds.x;
			} else if (right < bounds.x + bounds.width) {
				right = bounds.x + bounds.width;
			}
			if (top > bounds.y) {
				top = bounds.y;
			} else if (bottom < bounds.y + bounds.height) {
				bottom = bounds.y + bounds.height;
			}
		}

		// `̈쐬
		Rectangle selection = new Rectangle(left, top, right - left, bottom
				- top);
		selection.translate(getX(), getY());// y[W̍Wɖ߂

		// `̈Ԃ
		return selection;
	}

	/***************************************************************************
	 * GroupEditor ̎
	 **************************************************************************/

	public Group getGroup() {
		return this.textGroup;
	}

	/**
	 * AJ[ʒu߂B<br>
	 * iLbg[̉ʒu߂Bj <br>
	 * 
	 * @return AJ[ʒu
	 */
	public Point getAnchorPoint() {
		int caretOffset = getCaretPosition();
		Rectangle caretBounds;
		try {
			caretBounds = modelToView(caretOffset);
		} catch (BadLocationException e) {
			throw new RuntimeException("Lbg̃ItZbgsł");
		}
		localAnchorPoint.setLocation(caretBounds.x, caretBounds.y
				+ caretBounds.height);

		// O[vWɕϊ
		localAnchorPoint.translate(getX() - textGroup.getX(), getY()
				- textGroup.getY());
		return localAnchorPoint;
	}

	public void clearSelection() {
		select(0, 0);
	}

	public boolean contains(int x, int y) {
		return super.contains(x, y);
	}

	public boolean intersects(int x, int y, int width, int height) {
		Rectangle bounds = super.getBounds();
		return bounds.intersects(x, y, width, height);
	}

	/**
	 * @return I͈͂邩
	 */
	public boolean hasSelection() {
		int begin = getSelectionStart();
		int end = getSelectionEnd();
		return (end - begin) != 0;
	}

	public Object getSelection() {
		// TODO ꂽ\bhEX^u
		return null;
	}

	public void insertElement(NoteElement element) {
		// TODO ꂽ\bhEX^u

	}

	/**
	 * 
	 * @param group XVO[v
	 */
	public void setEditable(boolean isEditable) {
		// yC̕ҏWԂύX
		super.setEditable(isEditable);

		// ҏW\ԂɂȂtH[JXv
		if (isEditable) {
			requestFocus();
		}

		// pageUIɕҏW\ɂȂʒm
		// RXgNgɂ̃\bhĂ΂ꂽꍇApageUI == nullɂȂ̂ŉ
		if (pageUI != null) {
			pageUI.groupEditableChanged(this);
		}

	}

	public boolean isEditable() {
		return super.isEditable();
	}

	/***************************************************************************
	 * GroupChangedListener ̎
	 **************************************************************************/

	/**
	 * @param group XVO[v
	 */
	public void changedUpdate(Group group) {
		// GfB^̃TCYK؂ɐݒ肷
		setSuitableSize();
		setLocationToFirstLineBottom();
		pageUI.groupChanged(this);
	}

	/**
	 * @param group XVO[v
	 */
	public void movedUpdate(Group group) {
		setLocationToFirstLineBottom();
		pageUI.groupMoved(this);
	}

	/***************************************************************************
	 * JTextPane ̃I[o[Ch
	 **************************************************************************/
	/**
	 * ͂Ƃ̑Ԃ
	 * 
	 * @return ͑ۂ̑
	 * @see JTextPane#getInputAttributes()
	 */
	public MutableAttributeSet getInputAttributes() {
		// ݂̓͑iyɂĕύXꂽjԂ
		return TextGroup.getAttributeWithPen(currentPen);
	}

	/**
	 * @return Lbg̐F
	 */
	public Color getCaretColor() {
		if (currentPen != null) {
			return currentPen.getColor();
		} else {// RXgNgɌĂ΂ꂽ
			return Color.BLACK;
		}
	}

	/***************************************************************************
	 * `֘A
	 **************************************************************************/
	private class TextGroupPaneUI extends BasicTextPaneUI {

		/**
		 * UI`B
		 * 
		 * @see javax.swing.plaf.basic.BasicTextUI#paintSafely(java.awt.Graphics)
		 */
		// protected void paintSafely(Graphics g) {
		// ϊ̃eLXg̑i͋ZŁjς͕ύXȎ\Ȃ悤ɕ`sȂB
		// TODO ߂B̕@ƁAB
		// if (isChangedInputMethodTextAttributes) {
		// return;
		// } else {
		// super.paintSafely(g);
		// }
		// }
	}

	/**
	 * ̃R|[lg̕`揈B
	 * 
	 * @param g Graphics
	 */
	public void paintComponent(Graphics g) {
		super.paintComponent(g);

		// ҏW\Ԃ̏ꍇ̓O[v̍ɃKChC`悷
		if (isEditable()) {
			Dimension lineRec = getSize();
			if (lineRec != null) {
				g.setColor(Color.BLUE);
				g.drawLine(0, 0, 0, lineRec.height);
			}
		}
	}

	/*
	 * ItZbgʒu0hbgڂ̉[eLXgO[v̌_ɂȂ悤ɈړB
	 */
	private void setLocationToFirstLineBottom() {
		Rectangle firstOffsetBounds = getFirstOffsetBounds();
		int caretBottomY = firstOffsetBounds.y + firstOffsetBounds.height;// 0hbgڂ̉[yW
		int caretBottomX = firstOffsetBounds.x;// 0hbgڂ̉[xW
		int y = textGroup.getY() - caretBottomY;// eLXgO[v̈ړ
		int x = textGroup.getX() - caretBottomX;

		setLocation(x, y);
	}

	/*
	 * 0hbgڂ̃ItZbg̋`̈߂B
	 */
	private Rectangle getFirstOffsetBounds() {

		// 0ڂ̍߂
		try {
			Rectangle offsetBounds = modelToView(0);// 0Ԗڂ̃ItZbg̋`̈
			return offsetBounds;

		} catch (BadLocationException ex) {
			throw new RuntimeException("Psڂ̍܂܂" + ex.toString());
		}
	}

	/*
	 * eLXge\̂ɓKTCYɃGfB^[TCY
	 */
	private void setSuitableSize() {
		// ܂ԂKTCYԂȂ悤ɂƂ肠TCYw肷B
		// Ȃ킩ȂAgetPreferredSize()TCYԂĂȂ̂őËB
		// 1xKTCYw肷getPreferredSize()̒lXV悤B
		// TODO TAB͂ɂȂBjava ver1.4
		Dimension suitableSize = getPreferredSize();// KȃTCY
		setSize(suitableSize);

		// GfB^̗vKTCYɃGfB^TCY
		suitableSize = getPreferredSize();// KȃTCY
		setSize(suitableSize);
	}

	/**
	 * Lbgړ̏B<br>
	 * Lbgɍ킹ăAJ[|CgړB <br>
	 * 
	 * @param e CaretEvent
	 */
	public void caretUpdate(CaretEvent e) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				// LbgʒuɃAJ[ړ
				moveAnchorPointWithCaret();

				// AJ[̈ړʒm
				pageUI.groupAnchorPointMoved(TextGroupPane.this);
			}
		});
	}

	/*
	 * Lbgɍ킹ăAJ[|Cgړ @param dot Lbg̈ʒu
	 */
	private void moveAnchorPointWithCaret() {
		int dot = getCaretPosition();
		try {
			// Lbg̉ʒu擾
			Rectangle r = modelToView(dot);

			// y[W̍Wɒ
			int x = r.x;
			int y = r.y + r.height;
			localAnchorPoint.setLocation(x, y);

		} catch (Exception ex) {
			throw new RuntimeException("Lbg̃ItZbgs");
		}
	}

	/***************************************************************************
	 * I֘A
	 **************************************************************************/

	// /**
	// * @return IĂvf
	// */
	// public AbstractElement getSelectedElement() {
	// StyledDocument document = getStyledDocument();
	// Position begin;
	// Position end;
	// try {
	// begin = document.createPosition(getSelectionStart());
	// end = document.createPosition(getSelectionEnd());
	// } catch (BadLocationException e) {
	// e.printStackTrace();
	// throw new RuntimeException("I͈͂s");
	// }
	//
	// TextElement element = new TextElement(document, begin, end);
	// element.setParentGroup(textGroup);
	// return element;
	// }
	/***************************************************************************
	 * CxgnhO֘A
	 **************************************************************************/
	/**
	 * @see java.awt.Component#processMouseEvent(java.awt.event.MouseEvent)
	 */
	// protected void processMouseEvent(MouseEvent e) {
	//
	// // ҏWs\\
	// if (textGroup.isEditable()) {
	// super.processMouseEvent(e);
	//
	// } else {
	// MouseListener[] mouseListeners = getMouseListeners();
	// for (int i = 0; i < mouseListeners.length; i++) {
	// if (mouseListeners[i] instanceof TextGroupPaneHandler) {
	// int id = e.getID();
	// MouseListener l = mouseListeners[i];
	// switch (id) {
	// case MouseEvent.MOUSE_CLICKED:
	// l.mouseClicked(e);
	// break;
	// case MouseEvent.MOUSE_PRESSED:
	// l.mousePressed(e);
	// break;
	// case MouseEvent.MOUSE_RELEASED:
	// l.mouseReleased(e);
	// break;
	// }
	// }
	// }
	// }
	// }
	// /**
	// * @see
	// java.awt.Component#processMouseMotionEvent(java.awt.event.MouseEvent)
	// */
	// protected void processMouseMotionEvent(MouseEvent e) {
	// if (textGroup.isEditable()) {
	// super.processMouseMotionEvent(e);
	// } else {
	// MouseMotionListener[] mouseMotionListeners = getMouseMotionListeners();
	// for (int i = 0; i < mouseMotionListeners.length; i++) {
	// if (mouseMotionListeners[i] instanceof TextGroupPaneHandler) {
	// int id = e.getID();
	// MouseMotionListener l = mouseMotionListeners[i];
	// switch (id) {
	// case MouseEvent.MOUSE_DRAGGED:
	// l.mouseDragged(e);
	// break;
	// }
	// }
	// }
	// }
	// }
	/***************************************************************************
	 * InputMethod ֘A
	 **************************************************************************/
	private boolean isEnableInputMethod = false;// ϊ
	private boolean isChangedInputMethodTextAttributes = false;

	/**
	 * ϊǂĎ郊Xi[
	 */
	private class InputMethodWactcher implements InputMethodListener {

		/**
		 * @see java.awt.event.InputMethodListener#caretPositionChanged(java.awt.event.InputMethodEvent)
		 */
		public void caretPositionChanged(InputMethodEvent event) {
		}

		/**
		 * @see java.awt.event.InputMethodListener#inputMethodTextChanged(java.awt.event.InputMethodEvent)
		 */
		public void inputMethodTextChanged(InputMethodEvent event) {
			// ϊĨ^C~O𒲂ׂ
			if (event.getText() == null) {// meLXg+ϊeLXgȂꍇ
				// ϊI
				isEnableInputMethod = false;
			} else {
				isEnableInputMethod = true;
			}
		}
	}

	/**
	 * ϊ̕y̑ɕςNXB<br>
	 * ͋ZB
	 */
	private class DocumentAttributesChanger implements DocumentListener {

		public void insertUpdate(DocumentEvent e) {

			if (isEnableInputMethod) {// ϊ̏ꍇ
				final StyledDocument document = (StyledDocument) e
						.getDocument();
				final int begin = e.getOffset();
				final int length = e.getLength();

				isChangedInputMethodTextAttributes = true;
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {

						changeInputMethodTextAttributes(document, begin, length);
						isChangedInputMethodTextAttributes = false;

					}
				});
			}
		}

		public void removeUpdate(DocumentEvent e) {
		}

		public void changedUpdate(DocumentEvent e) {
		}

		private void changeInputMethodTextAttributes(StyledDocument document,
				int begin, int length) {
			document.setCharacterAttributes(begin, length, TextGroup
					.getAttributeWithPen(currentPen), false);
		}
	}

	/***************************************************************************
	 * Lbg
	 **************************************************************************/
	private class TextGroupCaret extends DefaultCaret {

		/***********************************************************************
		 * 萔
		 **********************************************************************/
		private static final long serialVersionUID = 1L;

		private Pen pen;

		public TextGroupCaret(Pen pen) {
			super();
			this.pen = pen;
		}

		public void paint(Graphics g) {
			super.paint(g);
		}

		public boolean isVisible() {
			// O[vҏW̓Lbg\Ȃ
			return pageUI.getMode().isGroupEditMode() || super.isVisible();
		}

	}

	/***************************************************************************
	 * Getter & Setter
	 **************************************************************************/
	// TODO 
	public TextGroup getTextGroup() {
		return this.textGroup;
	}

	/**
	 * @return InputMethodɂϊ
	 */
	public boolean isEnableInputMethod() {
		return isEnableInputMethod;
	}

}
